@@ -24,12 +24,213 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
 
-(* $Id: postgresql.ml,v 1.3 2004/06/29 21:14:05 mottl Exp $ *)
+(* $Id: postgresql.ml,v 1.5 2004/08/05 19:00:55 mottl Exp $ *)
 
 open Printf
 
+type oid = int
+type large_object = int
+
+exception Oid of oid
+exception InternalError of string
+
+let invalid_oid = 0
+
+type ftype =
+  | BOOL
+  | BYTEA
+  | CHAR
+  | NAME
+  | INT8
+  | INT2
+  | INT2VECTOR
+  | INT4
+  | REGPROC
+  | TEXT
+  | OID
+  | TID
+  | XID
+  | CID
+  | OIDVECTOR
+  | POINT
+  | LSEG
+  | PATH
+  | BOX
+  | POLYGON
+  | LINE
+  | FLOAT4
+  | FLOAT8
+  | ABSTIME
+  | RELTIME
+  | TINTERVAL
+  | UNKNOWN
+  | CIRCLE
+  | CASH
+  | MACADDR
+  | INET
+  | CIDR
+  | ACLITEM
+  | BPCHAR
+  | VARCHAR
+  | DATE
+  | TIME
+  | TIMESTAMP
+  | TIMESTAMPTZ
+  | INTERVAL
+  | TIMETZ
+  | BIT
+  | VARBIT
+  | NUMERIC
+  | REFCURSOR
+  | REGPROCEDURE
+  | REGOPER
+  | REGOPERATOR
+  | REGCLASS
+  | REGTYPE
+  | RECORD
+  | CSTRING
+  | ANY
+  | ANYARRAY
+  | VOID
+  | TRIGGER
+  | LANGUAGE_HANDLER
+  | INTERNAL
+  | OPAQUE
+  | ANYELEMENT
+
+external ftype_of_oid : oid -> ftype = "ftype_of_oid_stub"
+external oid_of_ftype : ftype -> oid = "oid_of_ftype_stub"
+
+let string_of_ftype = function
+  | BOOL -> "BOOL"
+  | BYTEA -> "BYTEA"
+  | CHAR -> "CHAR"
+  | NAME -> "NAME"
+  | INT8 -> "INT8"
+  | INT2 -> "INT2"
+  | INT2VECTOR -> "INT2VECTOR"
+  | INT4 -> "INT4"
+  | REGPROC -> "REGPROC"
+  | TEXT -> "TEXT"
+  | OID -> "OID"
+  | TID -> "TID"
+  | XID -> "XID"
+  | CID -> "CID"
+  | OIDVECTOR -> "OIDVECTOR"
+  | POINT -> "POINT"
+  | LSEG -> "LSEG"
+  | PATH -> "PATH"
+  | BOX -> "BOX"
+  | POLYGON -> "POLYGON"
+  | LINE -> "LINE"
+  | FLOAT4 -> "FLOAT4"
+  | FLOAT8 -> "FLOAT8"
+  | ABSTIME -> "ABSTIME"
+  | RELTIME -> "RELTIME"
+  | TINTERVAL -> "TINTERVAL"
+  | UNKNOWN -> "UNKNOWN"
+  | CIRCLE -> "CIRCLE"
+  | CASH -> "CASH"
+  | MACADDR -> "MACADDR"
+  | INET -> "INET"
+  | CIDR -> "CIDR"
+  | ACLITEM -> "ACLITEM"
+  | BPCHAR -> "BPCHAR"
+  | VARCHAR -> "VARCHAR"
+  | DATE -> "DATE"
+  | TIME -> "TIME"
+  | TIMESTAMP -> "TIMESTAMP"
+  | TIMESTAMPTZ -> "TIMESTAMPTZ"
+  | INTERVAL -> "INTERVAL"
+  | TIMETZ -> "TIMETZ"
+  | BIT -> "BIT"
+  | VARBIT -> "VARBIT"
+  | NUMERIC -> "NUMERIC"
+  | REFCURSOR -> "REFCURSOR"
+  | REGPROCEDURE -> "REGPROCEDURE"
+  | REGOPER -> "REGOPER"
+  | REGOPERATOR -> "REGOPERATOR"
+  | REGCLASS -> "REGCLASS"
+  | REGTYPE -> "REGTYPE"
+  | RECORD -> "RECORD"
+  | CSTRING -> "CSTRING"
+  | ANY -> "ANY"
+  | ANYARRAY -> "ANYARRAY"
+  | VOID -> "VOID"
+  | TRIGGER -> "TRIGGER"
+  | LANGUAGE_HANDLER -> "LANGUAGE_HANDLER"
+  | INTERNAL -> "INTERNAL"
+  | OPAQUE -> "OPAQUE"
+  | ANYELEMENT -> "ANYELEMENT"
+
+let ftype_of_string = function
+  | "BOOL" -> BOOL
+  | "BYTEA" -> BYTEA
+  | "CHAR" -> CHAR
+  | "NAME" -> NAME
+  | "INT8" -> INT8
+  | "INT2" -> INT2
+  | "INT2VECTOR" -> INT2VECTOR
+  | "INT4" -> INT4
+  | "REGPROC" -> REGPROC
+  | "TEXT" -> TEXT
+  | "OID" -> OID
+  | "TID" -> TID
+  | "XID" -> XID
+  | "CID" -> CID
+  | "OIDVECTOR" -> OIDVECTOR
+  | "POINT" -> POINT
+  | "LSEG" -> LSEG
+  | "PATH" -> PATH
+  | "BOX" -> BOX
+  | "POLYGON" -> POLYGON
+  | "LINE" -> LINE
+  | "FLOAT4" -> FLOAT4
+  | "FLOAT8" -> FLOAT8
+  | "ABSTIME" -> ABSTIME
+  | "RELTIME" -> RELTIME
+  | "TINTERVAL" -> TINTERVAL
+  | "UNKNOWN" -> UNKNOWN
+  | "CIRCLE" -> CIRCLE
+  | "CASH" -> CASH
+  | "MACADDR" -> MACADDR
+  | "INET" -> INET
+  | "CIDR" -> CIDR
+  | "ACLITEM" -> ACLITEM
+  | "BPCHAR" -> BPCHAR
+  | "VARCHAR" -> VARCHAR
+  | "DATE" -> DATE
+  | "TIME" -> TIME
+  | "TIMESTAMP" -> TIMESTAMP
+  | "TIMESTAMPTZ" -> TIMESTAMPTZ
+  | "INTERVAL" -> INTERVAL
+  | "TIMETZ" -> TIMETZ
+  | "BIT" -> BIT
+  | "VARBIT" -> VARBIT
+  | "NUMERIC" -> NUMERIC
+  | "REFCURSOR" -> REFCURSOR
+  | "REGPROCEDURE" -> REGPROCEDURE
+  | "REGOPER" -> REGOPER
+  | "REGOPERATOR" -> REGOPERATOR
+  | "REGCLASS" -> REGCLASS
+  | "REGTYPE" -> REGTYPE
+  | "RECORD" -> RECORD
+  | "CSTRING" -> CSTRING
+  | "ANY" -> ANY
+  | "ANYARRAY" -> ANYARRAY
+  | "VOID" -> VOID
+  | "TRIGGER" -> TRIGGER
+  | "LANGUAGE_HANDLER" -> LANGUAGE_HANDLER
+  | "INTERNAL" -> INTERNAL
+  | "OPAQUE" -> OPAQUE
+  | "ANYELEMENT" -> ANYELEMENT
+  | str -> failwith ("ftype_of_string: unknown ftype: " ^ str)
+
 external init : unit -> unit = "PQocaml_init"
-let _ = init ()
+let () =
+  Callback.register_exception "Postgresql.Oid" (Oid invalid_oid);
+  Callback.register_exception "Postgresql.InternalError" (InternalError "");
+  init ()
 
 type connection_status = Ok | Bad
 
@@ -64,11 +265,6 @@ type getline_async_result =
 
 external result_status : result_status -> string = "PQresStatus_stub"
 
-type oid = int
-type large_object = int
-
-let invalid_oid = 0
-
 type error =
   | Field_out_of_range of int * int
   | Tuple_out_of_range of int * int
@@ -100,7 +296,7 @@ module Stub = struct
   external conn_isnull : connection -> bool = "PQconn_isnull" "noalloc"
   external connect : string -> connection = "PQconnectdb_stub"
   external finish : connection -> unit = "PQfinish_stub"
-  external reset : connection -> unit = "PQreset_stub" "noalloc"
+  external reset : connection -> unit = "PQreset_stub"
 
   external db : connection -> string = "PQdb_stub"
   external user : connection -> string = "PQuser_stub"
@@ -281,7 +477,10 @@ object
     let n = Stub.fnumber res s in
     if n = -1 then raise Not_found else n
 
-  method ftype field = check_field field; Stub.ftype res field
+  method ftype field =
+    check_field field;
+    ftype_of_oid (Stub.ftype res field)
+
   method fmod field = check_field field; Stub.fmod res field
   method fsize field = check_field field; Stub.fsize res field
 
@@ -373,7 +572,7 @@ class connection ?host ?hostaddr ?port ?dbname ?user ?password ?options ?tty
 
   fun () ->
     let conn = Stub.connect conn_info in
-    let _ =
+    let () =
       if Stub.connection_status conn <> Ok then (
         let s = Stub.error_message conn in
         Stub.finish conn;
