@@ -24,7 +24,7 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-/* $Id: postgresql_stubs.c,v 1.5 2004/04/13 21:46:20 mottl Exp $ */
+/* $Id: postgresql_stubs.c,v 1.10 2004/12/21 03:41:42 mottl Exp $ */
 
 #include <string.h>
 
@@ -36,7 +36,68 @@
 #include <caml/fail.h>
 
 #include <libpq-fe.h>
-#include "libpq/libpq-fs.h"
+#include <libpq/libpq-fs.h>
+
+#define BOOLOID                   16
+#define BYTEAOID                  17
+#define CHAROID                   18
+#define NAMEOID                   19
+#define INT8OID                   20
+#define INT2OID                   21
+#define INT2VECTOROID             22
+#define INT4OID                   23
+#define REGPROCOID                24
+#define TEXTOID                   25
+#define OIDOID                    26
+#define TIDOID                    27
+#define XIDOID                    28
+#define CIDOID                    29
+#define OIDVECTOROID              30
+#define POINTOID                 600
+#define LSEGOID                  601
+#define PATHOID                  602
+#define BOXOID                   603
+#define POLYGONOID               604
+#define LINEOID                  628
+#define FLOAT4OID                700
+#define FLOAT8OID                701
+#define ABSTIMEOID               702
+#define RELTIMEOID               703
+#define TINTERVALOID             704
+#define UNKNOWNOID               705
+#define CIRCLEOID                718
+#define CASHOID                  790
+#define MACADDROID               829
+#define INETOID                  869
+#define CIDROID                  650
+#define ACLITEMOID              1033
+#define BPCHAROID               1042
+#define VARCHAROID              1043
+#define DATEOID                 1082
+#define TIMEOID                 1083
+#define TIMESTAMPOID            1114
+#define TIMESTAMPTZOID          1184
+#define INTERVALOID             1186
+#define TIMETZOID               1266
+#define BITOID                  1560
+#define VARBITOID               1562
+#define NUMERICOID              1700
+#define REFCURSOROID            1790
+#define REGPROCEDUREOID         2202
+#define REGOPEROID              2203
+#define REGOPERATOROID          2204
+#define REGCLASSOID             2205
+#define REGTYPEOID              2206
+#define RECORDOID               2249
+#define CSTRINGOID              2275
+#define ANYOID                  2276
+#define ANYARRAYOID             2277
+#define VOIDOID                 2278
+#define TRIGGEROID              2279
+#define LANGUAGE_HANDLEROID     2280
+#define INTERNALOID             2281
+#define OPAQUEOID               2282
+#define ANYELEMENTOID           2283
 
 static value empty_string = 0;
 static value none = Val_int(0);
@@ -50,13 +111,48 @@ static value make_some(value v)
   CAMLreturn(res);
 }
 
+/* Cache for exceptions */
+static value *exc_Oid = NULL;  /* Exception [Oid] */
+static value *exc_InternalError = NULL;  /* Exception [InternalError] */
+
 CAMLprim value PQocaml_init()
 {
   register_global_root(&empty_string);
   empty_string = alloc_string(0);
+  exc_Oid = caml_named_value("Postgresql.Oid");
+  exc_InternalError = caml_named_value("Postgresql.InternalError");
   return Val_unit;
 }
 
+
+/* Conversion functions */
+
+static int oid_tbl[] = {
+  BOOLOID, BYTEAOID, CHAROID, NAMEOID, INT8OID, INT2OID, INT2VECTOROID, INT4OID,
+  REGPROCOID, TEXTOID, OIDOID, TIDOID, XIDOID, CIDOID, OIDVECTOROID, POINTOID,
+  LSEGOID, PATHOID, BOXOID, POLYGONOID, LINEOID, FLOAT4OID, FLOAT8OID, ABSTIMEOID,
+  RELTIMEOID, TINTERVALOID, UNKNOWNOID, CIRCLEOID, CASHOID, MACADDROID, INETOID,
+  CIDROID, ACLITEMOID, BPCHAROID, VARCHAROID, DATEOID, TIMEOID, TIMESTAMPOID,
+  TIMESTAMPTZOID, INTERVALOID, TIMETZOID, BITOID, VARBITOID, NUMERICOID,
+  REFCURSOROID, REGPROCEDUREOID, REGOPEROID, REGOPERATOROID, REGCLASSOID,
+  REGTYPEOID, RECORDOID, CSTRINGOID, ANYOID, ANYARRAYOID, VOIDOID, TRIGGEROID,
+  LANGUAGE_HANDLEROID, INTERNALOID, OPAQUEOID, ANYELEMENTOID
+};
+
+CAMLprim value ftype_of_oid_stub(value v_oid) {
+  int oid = Int_val(v_oid);
+  int *p = oid_tbl;
+  int *last = oid_tbl + sizeof(oid_tbl)/sizeof(oid_tbl[0]);
+  while(p != last && *p != oid) p++;
+  if (p == last) caml_raise_with_arg(*exc_Oid, v_oid);
+  return Val_int (p - oid_tbl);
+}
+
+CAMLprim value oid_of_ftype_stub(value v_ftype) {
+  return Val_int(oid_tbl[Int_val(v_ftype)]);
+}
+
+
 /* Management of notice_processor callbacks */
 
 /* One must me careful with notice processors: the callback
@@ -121,16 +217,18 @@ static void free_conn(value vconn)
   set_conn(vconn, NULL);
 }
 
-CAMLprim value PQconnectdb_stub(value conn_info)
+CAMLprim value PQconnectdb_stub(value vconn_info)
 {
   PGconn *conn;
   value vconn;
 
-  Begin_root(conn_info);
-  enter_blocking_section();
-    conn = PQconnectdb(String_val(conn_info));
-  leave_blocking_section();
-  End_roots();
+  char *conn_info = malloc((string_length(vconn_info) + 1));
+  strcpy(conn_info, String_val(vconn_info));
+
+  caml_enter_blocking_section();
+    conn = PQconnectdb(conn_info);
+    free(conn_info);
+  caml_leave_blocking_section();
 
   /* One may raise this 30 to 500 for instance if the program takes
      responsability of closing connections */
@@ -151,10 +249,10 @@ CAMLprim value PQfinish_stub(value vconn)
 CAMLprim value PQreset_stub(value vconn)
 {
   CAMLparam1(vconn);
-  enter_blocking_section();
+  caml_enter_blocking_section();
     PQreset(get_conn(vconn));
-  leave_blocking_section();
-  CAMLreturn (Val_unit);
+  caml_leave_blocking_section();
+  CAMLreturn(Val_unit);
 }
 
 CAMLprim value PQconndefaults_stub()
@@ -370,7 +468,7 @@ CAMLprim value PQescapeBytea_stub(value v_from, value v_pos_from, value v_len)
     (char *) PQescapeBytea(
       (unsigned char *) String_val(v_from) + Int_val(v_pos_from),
       Int_val(v_len), &len);
-  value v_res = alloc_string(len);
+  value v_res = alloc_string(len - 1);
   memcpy(String_val(v_res), buf, len);
   free(buf);
   return v_res;
